@@ -28,6 +28,8 @@
 #include <regex.h>
 #include <math.h>
 
+#define PANGO_ENABLE_BACKEND /* Required to get PANGO_GLYPH_EMPTY */
+
 #include <pango/pango.h>
 
 #include "htmltext.h"
@@ -53,6 +55,12 @@ static const PangoAttrClass html_pango_attr_font_size_klass;
 
 #define HT_CLASS(x) HTML_TEXT_CLASS (HTML_OBJECT (x)->klass)
 
+#ifdef PANGO_GLYPH_EMPTY
+#define EMPTY_GLYPH PANGO_GLYPH_EMPTY
+#else
+#define EMPTY_GLYPH 0
+#endif
+
 static SpellError * spell_error_new         (guint off, guint len);
 static void         spell_error_destroy     (SpellError *se);
 static void         move_spell_errors       (GList *spell_errors, guint offset, gint delta);
@@ -898,6 +906,8 @@ html_text_get_attr_list_list (PangoAttrList *get_attrs, PangoAttrList *attr_list
 				g_slist_free (l);
 			}
 		} while (pango_attr_iterator_next (iter));
+
+		pango_attr_iterator_destroy (iter);
 	}
 }
 
@@ -936,8 +946,11 @@ html_text_calc_part_width (HTMLText *text, HTMLPainter *painter, char *start, gi
 	PangoFont *font = NULL;
 	gchar *s = start;
 
-	g_return_val_if_fail (offset >= 0, 0);
-	g_return_val_if_fail (offset + len <= text->text_len, 0);
+	if (offset < 0)
+		return 0;
+
+	if (offset + len > text->text_len)
+		return 0;
 
 	if (need_ascent_descent) {
 		ascent = html_painter_engine_to_pango (painter,
@@ -1173,15 +1186,21 @@ html_text_remove_unwanted_line_breaks (char *s, int len, PangoLogAttr *attrs)
 	}
 }
 
-static PangoAttrList *
-prepare_attrs (HTMLText *text, HTMLPainter *painter)
+PangoAttrList *
+html_text_prepare_attrs (HTMLText *text, HTMLPainter *painter)
 {
 	PangoAttrList *attrs;
 	HTMLClueFlow *flow = NULL;
 	HTMLEngine *e = NULL;
 	PangoAttribute *attr;
+	PangoDirection dir,org_dir;
 
 	attrs = pango_attr_list_new ();
+	org_dir = pango_context_get_base_dir (painter->pango_context);
+	dir = html_text_get_pango_direction (text);
+	if (org_dir != dir) 
+		pango_context_set_base_dir (painter->pango_context, dir);
+
 
 	if (HTML_OBJECT (text)->parent && HTML_IS_CLUEFLOW (HTML_OBJECT (text)->parent))
 		flow = HTML_CLUEFLOW (HTML_OBJECT (text)->parent);
@@ -1189,8 +1208,15 @@ prepare_attrs (HTMLText *text, HTMLPainter *painter)
 	if (painter->widget && GTK_IS_HTML (painter->widget))
 		e = html_object_engine (HTML_OBJECT (text), GTK_HTML (painter->widget)->engine);
 
-	if (flow && e)
+	if (flow && e) {
 		html_text_add_cite_color (attrs, text, flow, e);
+		if (org_dir != dir) {
+			html_engine_freeze (e);
+                	flow->dir = html_text_direction_pango_to_html (dir);
+                	html_engine_thaw(e);
+		}
+        }
+
 
 	if (HTML_IS_PLAIN_PAINTER (painter)) {
 		attr = pango_attr_family_new (painter->font_manager.fixed.face);
@@ -1221,14 +1247,18 @@ prepare_attrs (HTMLText *text, HTMLPainter *painter)
 	}
 
 	if (text->links && e) {
-		HTMLColor *link_color = html_colorset_get_color (e->settings->color_set, HTMLLinkColor);
+		HTMLColor *link_color;
 		GSList *l;
 
 		for (l = text->links; l; l = l->next) {
-			PangoAttribute *attr;
 			Link *link;
 
 			link = (Link *) l->data;
+
+			if(link->is_visited == FALSE) 
+				link_color = html_colorset_get_color (e->settings->color_set, HTMLLinkColor);
+			else 
+				link_color = html_colorset_get_color (e->settings->color_set, HTMLVLinkColor);
 			attr = pango_attr_underline_new (PANGO_UNDERLINE_SINGLE);
 			attr->start_index = link->start_index;
 			attr->end_index = link->end_index;
@@ -1351,7 +1381,7 @@ html_text_shape_tab (HTMLText *text, PangoGlyphString *glyphs)
 	/* copied from pango sources */
 	pango_glyph_string_set_size (glyphs, 1);
 
-	glyphs->glyphs[0].glyph = 0;
+	glyphs->glyphs[0].glyph = EMPTY_GLYPH;
 	glyphs->glyphs[0].geometry.x_offset = 0;
 	glyphs->glyphs[0].geometry.y_offset = 0;
 	glyphs->glyphs[0].attr.is_cluster_start = 1;
@@ -1374,7 +1404,7 @@ html_text_get_pango_info (HTMLText *text, HTMLPainter *painter)
 		PangoAttrList *attrs;
 		int i, offset;
 
-		attrs = prepare_attrs (text, painter);
+		attrs = html_text_prepare_attrs (text, painter);
 		items = pango_itemize_with_base_dir (painter->pango_context, get_pango_base_direction (text), text->text, 0, text->text_bytes, attrs, NULL);
 		pango_attr_list_unref (attrs);
 
@@ -1624,7 +1654,7 @@ save_open_attrs (HTMLEngineSaveState *state, GSList *attrs)
 		case PANGO_ATTR_FAMILY: {
 			PangoAttrString *family_attr = (PangoAttrString *) attr;
 
-			if (!strcasecmp (e->painter->font_manager.fixed.face
+			if (!g_ascii_strcasecmp (e->painter->font_manager.fixed.face
 					? e->painter->font_manager.fixed.face : "Monospace",
 					family_attr->value))
 				tag = "<TT>";
@@ -1688,7 +1718,7 @@ save_close_attrs (HTMLEngineSaveState *state, GSList *attrs)
 		case PANGO_ATTR_FAMILY: {
 			PangoAttrString *family_attr = (PangoAttrString *) attr;
 
-			if (!strcasecmp (e->painter->font_manager.fixed.face
+			if (!g_ascii_strcasecmp (e->painter->font_manager.fixed.face
 					? e->painter->font_manager.fixed.face : "Monospace",
 					family_attr->value))
 				tag = "</TT>";
@@ -1784,10 +1814,12 @@ save (HTMLObject *self, HTMLEngineSaveState *state)
 
 		do {
 			GSList *attrs;
-			guint start_index, end_index;
+			gint start_index, end_index;
 
 			attrs = pango_attr_iterator_get_attrs (iter);
-			pango_attr_iterator_range (iter, &start_index, &end_index);
+			pango_attr_iterator_range (iter,
+						   &start_index,
+						   &end_index);
 			if (end_index > text->text_bytes)
 				end_index = text->text_bytes;
 
@@ -1800,6 +1832,8 @@ save (HTMLObject *self, HTMLEngineSaveState *state)
 				html_text_free_attrs (attrs);
 			}
 		} while (pango_attr_iterator_next (iter));
+
+		pango_attr_iterator_destroy (iter);
 		g_slist_free (links);
 	}
 
@@ -1984,7 +2018,9 @@ convert_nbsp (gchar *fill, const gchar *text)
 }
 
 static void
-update_index_interval (int *start_index, int *end_index, GSList *changes)
+update_index_interval (guint *start_index,
+		       guint *end_index,
+		       GSList *changes)
 {
 	GSList *c;
 	int index, delta;
@@ -2108,7 +2144,10 @@ static GList *
 remove_one (GList *list, GList *link)
 {
 	spell_error_destroy ((SpellError *) link->data);
-	return g_list_remove_link (list, link);
+	list = g_list_remove_link (list, link);
+	g_list_free (link);
+
+	return list;
 }
 
 static GList *
@@ -2362,7 +2401,9 @@ append_selection_string (HTMLObject *self,
 	last = html_text_get_text (text,
 				   text->select_start + text->select_length);
 	*/
-	html_engine_save_string_append_nonbsp (buffer, p, last - p);
+	html_engine_save_string_append_nonbsp (buffer,
+					       (guchar *)p,
+					       last - p);
 
 }
 
@@ -2675,10 +2716,12 @@ offset_to_pointer_validated (const gchar *str, glong offset, gint *chars_out)
 gsize
 html_text_sanitize (const gchar **str, gint *len)
 {
+	gchar *end;
+
 	g_return_val_if_fail (str != NULL, 0);
 	g_return_val_if_fail (len != NULL, 0);
 	
-	gchar *end = offset_to_pointer_validated (*str, *len, len);
+	end = offset_to_pointer_validated (*str, *len, len);
 	if (end) {
 		return end - *str;
 	} else {
@@ -2783,7 +2826,8 @@ html_text_set_text (HTMLText *text, const gchar *new_text)
 {
 	g_free (text->text);
 	text->text_len = -1;
-	text->text_bytes = html_text_sanitize (&new_text, &text->text_len);
+	text->text_bytes = html_text_sanitize (&new_text,
+					       (gint *)&text->text_len);
 	text->text = g_memdup (new_text, text->text_bytes + 1);
 	text->text [text->text_bytes] = '\0';
 	html_object_change_set (HTML_OBJECT (text), HTML_CHANGE_ALL);
@@ -2919,9 +2963,9 @@ struct _HTMLMagicInsertMatch
 typedef struct _HTMLMagicInsertMatch HTMLMagicInsertMatch;
 
 static HTMLMagicInsertMatch mim [] = {
-	{ "(news|telnet|nttp|file|http|ftp|https)://([-a-z0-9]+(:[-a-z0-9]+)?@)?[-a-z0-9.]+[-a-z0-9](:[0-9]*)?(/[-a-z0-9_$.+!*(),;:@%&=?/~#]*[^]'.}>\\) ,?!;:\"]?)?", NULL, NULL },
-	{ "www\\.[-a-z0-9.]+[-a-z0-9](:[0-9]*)?(/[-A-Za-z0-9_$.+!*(),;:@%&=?/~#]*[^]'.}>\\) ,?!;:\"]?)?", NULL, "http://" },
-	{ "ftp\\.[-a-z0-9.]+[-a-z0-9](:[0-9]*)?(/[-A-Za-z0-9_$.+!*(),;:@%&=?/~#]*[^]'.}>\\) ,?!;:\"]?)?", NULL, "ftp://" },
+	{ "(news|telnet|nttp|file|http|ftp|sftp|https)://([-a-z0-9]+(:[-a-z0-9]+)?@)?[-a-z0-9.]+[-a-z0-9](:[0-9]*)?(([.])?/[-a-z0-9_$.+!*(),;:@%&=?/~#']*[^]'.}>\\) ,?!;:\"]?)?", NULL, NULL },
+	{ "www\\.[-a-z0-9.]+[-a-z0-9](:[0-9]*)?(([.])?/[-A-Za-z0-9_$.+!*(),;:@%&=?/~#]*[^]'.}>\\) ,?!;:\"]?)?", NULL, "http://" },
+	{ "ftp\\.[-a-z0-9.]+[-a-z0-9](:[0-9]*)?(([.])?/[-A-Za-z0-9_$.+!*(),;:@%&=?/~#]*[^]'.}>\\) ,?!;:\"]?)?", NULL, "ftp://" },
 	{ "[-_a-z0-9.'\\+]+@[-_a-z0-9.]+", NULL, "mailto:" }
 };
 
@@ -3060,7 +3104,7 @@ html_text_append (HTMLText *text, const gchar *str, gint len)
 void
 html_text_append_link_full (HTMLText *text, gchar *url, gchar *target, gint start_index, gint end_index, gint start_offset, gint end_offset)
 {
-	text->links = g_slist_prepend (text->links, html_link_new (url, target, start_index, end_index, start_offset, end_offset));
+	text->links = g_slist_prepend (text->links, html_link_new (url, target, start_index, end_index, start_offset, end_offset, FALSE));
 }
 
 static void
@@ -3090,7 +3134,7 @@ html_text_add_link_full (HTMLText *text, HTMLEngine *e, gchar *url, gchar *targe
 	if (text->links == NULL)
 		html_text_append_link_full (text, url, target, start_index, end_index, start_offset, end_offset);
 	else {
-		Link *plink = NULL, *new_link = html_link_new (url, target, start_index, end_index, start_offset, end_offset);
+		Link *plink = NULL, *new_link = html_link_new (url, target, start_index, end_index, start_offset, end_offset, FALSE);
 
 		for (l = text->links; new_link && l; l = l->next) {
 			link = (Link *) l->data;
@@ -3306,6 +3350,7 @@ html_link_dup (Link *l)
 	nl->end_offset = l->end_offset;
 	nl->start_index = l->start_index;
 	nl->end_index = l->end_index;
+	nl->is_visited = l->is_visited;
 
 	return nl;
 }
@@ -3323,12 +3368,12 @@ html_link_free (Link *link)
 gboolean
 html_link_equal (Link *l1, Link *l2)
 {
-	return l1->url && l2->url && !strcasecmp (l1->url, l2->url)
-		&& (l1->target == l2->target || (l1->target && l2->target && !strcasecmp (l1->target, l2->target)));
+	return l1->url && l2->url && !g_ascii_strcasecmp (l1->url, l2->url)
+		&& (l1->target == l2->target || (l1->target && l2->target && !g_ascii_strcasecmp (l1->target, l2->target)));
 }
 
 Link *
-html_link_new (gchar *url, gchar *target, guint start_index, guint end_index, gint start_offset, gint end_offset)
+html_link_new (gchar *url, gchar *target, guint start_index, guint end_index, gint start_offset, gint end_offset, gboolean is_visited)
 {
 	Link *link = g_new0 (Link, 1);
 
@@ -3338,6 +3383,7 @@ html_link_new (gchar *url, gchar *target, guint start_index, guint end_index, gi
 	link->end_offset = end_offset;
 	link->start_index = start_index;
 	link->end_index = end_index;
+	link->is_visited = is_visited;
 
 	return link;
 }
@@ -3714,3 +3760,28 @@ html_text_direction_pango_to_html (PangoDirection pdir)
 		return HTML_DIRECTION_DERIVED;
 	}
 }
+
+void
+html_text_change_set (HTMLText *text, HTMLChangeFlags flags)
+{
+	HTMLObject *slave = HTML_OBJECT (text)->next;
+
+	for (; slave && HTML_IS_TEXT_SLAVE (slave) && HTML_TEXT_SLAVE (slave)->owner == text; slave = slave->next)
+		slave->change |= flags;
+
+	html_object_change_set (HTML_OBJECT (text), flags);
+}
+
+void
+html_text_set_link_visited (HTMLText *text, gint offset, HTMLEngine *engine, gboolean is_visited)
+{
+        HTMLEngine *object_engine=html_object_engine(HTML_OBJECT(text),engine); 
+        Link *link = html_text_get_link_at_offset (text,offset);
+
+	if (link) {
+                link->is_visited  = is_visited;
+                html_text_change_set (text, HTML_CHANGE_RECALC_PI);
+                html_text_queue_draw (text, object_engine, offset, 1);
+                html_engine_flush_draw_queue (object_engine);
+        }
+}
